///////////////////////////////////////////////////////////////////////////////
// Created on: April 2000
// Created by: Ronald Hordijk (hk@dgmr.nl)
//
// MainForm
//  - Handle user input
//  - File i/o with ini file.
//  - Regulate the calculator and paint threads.
///////////////////////////////////////////////////////////////////////////////
unit mainform;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Opengl12,
  ExtCtrls, Vector, StdCtrls, ComCtrls, RaEdit, Gauges, RangeSel, FaceList,
  Buttons, JuliaCalc, PaintThread, CalcThread, Coordinates;

type
  TfrmMain = class(TForm)
    StatusBar: TStatusBar;
    PageControl: TPageControl;
    Parameters: TTabSheet;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label10: TLabel;
    cmbType: TComboBox;
    RaeX: TRaEdit;
    RaeY: TRaEdit;
    raeZ: TRaEdit;
    raeW: TRaEdit;
    raeNrInter: TRaEdit;
    Dimensions: TTabSheet;
    rsX: TRangeSelector;
    Label9: TLabel;
    Label11: TLabel;
    rsY: TRangeSelector;
    Label12: TLabel;
    rsZ: TRangeSelector;
    raeXLow: TRaEdit;
    raeXHigh: TRaEdit;
    raeXLowNew: TRaEdit;
    raeXHighNew: TRaEdit;
    raeYLow: TRaEdit;
    raeYHigh: TRaEdit;
    raeYLowNew: TRaEdit;
    raeYHighNew: TRaEdit;
    raeZLow: TRaEdit;
    raeZHigh: TRaEdit;
    raeZLowNew: TRaEdit;
    raeZHighNew: TRaEdit;
    BBtnReset: TBitBtn;
    bbtnZoom: TBitBtn;
    Viewing: TTabSheet;
    btnRepaint: TBitBtn;
    BbtnViewSettingsRepaint: TBitBtn;
    raeGridSize: TRaEdit;
    Label1: TLabel;
    Label13: TLabel;
    pnlObjectColor: TPanel;
    raeQuickView: TRaEdit;
    Label14: TLabel;
    CbBoolBorder: TCheckBox;
    cbFog: TCheckBox;
    rbPoints: TRadioButton;
    rbLines: TRadioButton;
    rbPolygons: TRadioButton;
    rbLight: TRadioButton;
    Label15: TLabel;
    pnlBackColor: TPanel;
    BbtnRandom: TBitBtn;
    btnSave: TButton;
    Label2: TLabel;
    tbLoad: TTrackBar;
    rbTexturemap: TRadioButton;
    procedure FormCreate(Sender: TObject);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormResize(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure rsValueChanged(Sender: TObject);
    procedure raeLowNewChange(Sender: TObject);
    procedure raeHighNewChange(Sender: TObject);
    procedure BBtnResetClick(Sender: TObject);
    procedure bbtnZoomClick(Sender: TObject);
    procedure btnRepaintClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure pnlObjectColorClick(Sender: TObject);
    procedure PageControlChange(Sender: TObject);
    procedure cbFogClick(Sender: TObject);
    procedure rbLightClick(Sender: TObject);
    procedure pnlBackColorClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure BbtnRandomClick(Sender: TObject);
    procedure raeQuickViewChange(Sender: TObject);
    procedure tbLoadChange(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
  private
    { Private declarations }
    ParamEdit:      array[0..3] of TCustomEdit;
    DimEdit:        array[0..2,0..3] of TCustomEdit;
    RangeSelector:  array[0..2] of TRangeSelector;
    DrawModeRadio:  array[0..4] of TRadioButton;

    CalcParam:      TCalcParam;
    Coordinates:    TCoordinates;
    PaintParam:     TPaintParam;
    Dragging:       Boolean;
    CalcThread:     TCalcThread;
    FaceList:       TFaceList;
    MousePos:       TPoint;
    NrSaves:        Integer;
    iniFile:        TFilename;

    PT: TPaintThread;

    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;

    procedure FillControls;
    procedure SetDefaultStartStop;

    procedure Calculate;

    procedure GlInit;

    procedure RotateCamera(Right, Up, Turn: double);

    procedure ReadViewSettingsFromIni;
    procedure WriteViewSettingsToIni;
    procedure ReadScreenPosFromIni;
    procedure WriteScreenPosToIni;
    procedure GetNrSave;
    function CreateTexture(Texture: String): TGluInt;

  public
    { Public declarations }
  end;


var
  frmMain: TfrmMain;

implementation

uses
  inifiles, math;

{$R *.DFM}

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormCreate(Sender: TObject);
begin
  ParamEdit[0] := raeX;
  ParamEdit[1] := raeY;
  ParamEdit[2] := raeZ;
  ParamEdit[3] := raeW;

  DimEdit[0,0] := raeXLow;
  DimEdit[0,1] := raeXHigh;
  DimEdit[0,2] := raeXLowNew;
  DimEdit[0,3] := raeXHighNew;
  DimEdit[1,0] := raeYLow;
  DimEdit[1,1] := raeYHigh;
  DimEdit[1,2] := raeYLowNew;
  DimEdit[1,3] := raeYHighNew;
  DimEdit[2,0] := raeZLow;
  DimEdit[2,1] := raeZHigh;
  DimEdit[2,2] := raeZLowNew;
  DimEdit[2,3] := raeZHighNew;

  RangeSelector[0] := rsX;
  RangeSelector[1] := rsY;
  RangeSelector[2] := rsZ;

  DrawModeRadio[0] := rbPoints;
  DrawModeRadio[1] := rbLines;
  DrawModeRadio[2] := rbPolygons;
  DrawModeRadio[3] := rbLight;
  DrawModeRadio[4] := rbTexturemap;

  cmbType.Items.Add('Q-SQR');
  cmbType.Items.Add('Q-Cube');
  cmbType.Items.Add('H-SQR');
  cmbType.Items.Add('H-Cube');
  cmbType.Items.Add('H-Exp');
  cmbType.Items.Add('H-Sin');
  cmbType.Items.Add('H-Cos');
  cmbType.Items.Add('H-Sinh');
  cmbType.Items.Add('H-ASinh');
  cmbType.Items.Add('H-Cosh');
  cmbType.Items.Add('H-ACosh');
//  cmbType.Items.Add('H-1/h');
//  cmbType.Items.Add('H-Log');
//  cmbType.Items.Add('H-ASin');
//  cmbType.Items.Add('H-ACos');
//  cmbType.Items.Add('H-ATan');
//  cmbType.Items.Add('H-ATanh');
//  cmbType.Items.Add('H-Tan');
//  cmbType.Items.Add('H-Tanh');

  Coordinates := TCoordinates.Create;
  FaceList := TFaceList.Create;

  Randomize();

  IniFile :=  IncludeTrailingBackslash(ExtractFilePath(ParamStr(0))) + 'fract3d.ini';

  CalcParam.CalcType := 0;
  CalcParam.Border := False;
  PaintParam.Azimuth := 0;
  PaintParam.AngleHeight := 0;

  CalcParam.Param[0]:= -0.083;
  CalcParam.Param[1]:= 0;
  CalcParam.Param[2]:= -0.83;
  CalcParam.Param[3]:= -0.025;
  CalcParam.NrInter := 10;

  ReadViewSettingsFromIni;

  SetDefaultStartStop;

  PaintParam.ViewUp     := Vector3(0,0,1);
  PaintParam.ViewRight  := Vector3(0,1,0);
  PaintParam.ViewLook   := Vector3(1,0,0);

  Dragging := false;

  PageControl.ActivePageIndex := 0;

  GetNrSave;

  PT := nil;
  CalcThread := nil;
  GLInit;
  FillControls;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  if Assigned(CalcThread) then begin
    CalcThread.Terminate;
    CalcThread.WaitFor;
    FreeAndNil(CalcThread);
  end;

  if Assigned(PT) then begin
    PT.Terminate;
    PT.WaitFor;
    FreeAndNil(PT);
  end;

  wglMakeCurrent(0, 0);
  wglDeleteContext(PaintParam.RC);
  ReleaseDC(Handle, PaintParam.DC);

  WriteScreenPosToIni;

  FaceList.Free;
  Coordinates.Free;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.GlInit;
const
  glfLightAmbient : Array[0..3] of TGLfloat = (0.1, 0.1, 0.1, 1.0);
  glfLightDiffuse : Array[0..3] of TGLfloat = (0.7, 0.7, 0.7, 1.0);
  glfLightSpecular: Array[0..3] of TGLfloat = (1.0, 1.0, 1.0, 1.0);
  mat_specular    : Array[0..3] of TGLfloat = (1, 1, 1, 1.0);
  mat_shininess: TGLfloat   = 40;

  glfFog          : Array[0..3] of TGLfloat = (0.0, 0.0, 0.0, 1.0);
  glfLightPosition: Array[0..3] of TGLfloat = (0.0, 0.0, -100.0, 1.0);
begin
  initOpengl;

  PaintParam.DC := GetDC(Handle);
  PaintParam.RC := CreateRenderingContext(PaintParam.DC, [opDoubleBuffered], 16, 0, 0, 0, 0);

  wglMakeCurrent(PaintParam.DC, PaintParam.RC);

  glEnable(GL_DEPTH_TEST);

  glLightfv(GL_LIGHT1, GL_DIFFUSE , @glfLightDiffuse);
  glLightfv(GL_LIGHT1, GL_POSITION, @glfLightPosition);

  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);

  glFogi (GL_FOG_MODE, GL_LINEAR);
  glHint (GL_FOG_HINT, GL_NICEST);
  glFogf (GL_FOG_START, 5.0);
  glFogf (GL_FOG_END, 7.5);
  glFogfv (GL_FOG_COLOR, @glfFog);

  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glHint (GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
  glLineWidth (1.2);

  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
  glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);

//  Enable this line to use textures
//  glBindTexture(GL_TEXTURE_2D, CreateTexture('Texture.bmp'));

  wglMakeCurrent(0,0);
end;

///////////////////////////////////////////////////////////////////////////////
function TfrmMain.CreateTexture(Texture : String): TGluInt;
type
  PPixelArray = ^TPixelArray;
  TPixelArray = array [0..0] of Byte;
var
  bitmap: TBitmap;
  BMInfo : TBitmapInfo;
  ImageSize : Integer;
  MemDC : HDC;
  Tex: PPixelArray;
begin
  glGenTextures(1, @Result);
  glBindTexture(GL_TEXTURE_2D, Result);

  Bitmap := TBitMap.Create;

  Bitmap.LoadFromFile(Texture);
  with BMinfo.bmiHeader do begin
    FillChar (BMInfo, SizeOf(BMInfo), 0);
    biSize := sizeof (TBitmapInfoHeader);
    biBitCount := 24;
    biWidth := Bitmap.Width;
    biHeight := Bitmap.Height;
    ImageSize := biWidth * biHeight;
    biPlanes := 1;
    biCompression := BI_RGB;

    MemDC := CreateCompatibleDC (0);
    GetMem(Tex, ImageSize *3);
    try
      GetDIBits (MemDC, Bitmap.Handle, 0, biHeight, Tex, BMInfo, DIB_RGB_COLORS);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      gluBuild2DMipmaps(GL_TEXTURE_2D, 3, Bitmap.Width ,
                           Bitmap.Height, GL_BGR, GL_UNSIGNED_BYTE, Tex);
     finally
      DeleteDC (MemDC);
      Bitmap.Free;
      freemem(tex);
   end;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.Calculate;
begin
  if assigned(CalcThread) then begin
    CalcThread.Terminate;
    CalcThread.WaitFor;
    FreeAndNil(CalcThread);
  end;

  CalcThread := TCalcThread.Create(FaceList, Coordinates, CalcParam, Self);
  Invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormPaint(Sender: TObject);
begin
  if assigned(PT) then begin
    PT.Terminate;
    PT.WaitFor;
    FreeAndNil(PT);
  end;
  PaintParam.QuickDraw := Dragging or rsX.Dragging or rsY.Dragging or rsZ.Dragging;
  PaintParam.DrawBox := PageControl.ActivePageIndex = 1;
  PT := TPaintThread.Create(FaceList, PaintParam, Coordinates);
end;

///////////////////////////////////////////////////////////////////////////////
procedure TFrmMain.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormResize(Sender: TObject);
begin
  if assigned(PT) then begin
    PT.Terminate;
    PT.WaitFor;
    FreeAndNil(PT);
  end;

  wglMakeCurrent(PaintParam.DC, PaintParam.RC);
  glViewport(0 , 0, (ClientWidth - PageControl.Width), (ClientHeight - StatusBar.Height));
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(30, (ClientWidth - PageControl.Width)/(ClientHeight - StatusBar.Height), 4, 8);
  wglMakeCurrent(0,0);
  Refresh;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FillControls;
var
  i: integer;
begin
  cmbType.ItemIndex := CalcParam.CalcType;

  for i := 0 to 3 do
    ParamEdit[i].Text := Format('%.7f', [CalcParam.Param[i]]);

  raeNrInter.Text := IntToStr(CalcParam.NrInter);

  for i := 0 to 2 do begin
    DimEdit[i,0].Text := FloatToStr(Coordinates.Start[i]);
    DimEdit[i,1].Text := FloatToStr(Coordinates.Stop[i]);
    DimEdit[i,2].Text := DimEdit[i,0].Text;
    DimEdit[i,3].Text := DimEdit[i,1].Text;

    RangeSelector[i].LowValue := 0;
    RangeSelector[i].HighValue := 1000;
  end;

  raeGridSize.Text := IntToStr(Coordinates.GridSize);
  raeQuickView.Text := IntToStr(PaintParam.QuickDrawSize);

  pnlObjectColor.Color := PaintParam.ObjectColor;
  pnlBackColor.Color := PaintParam.BackColor;

  cbBoolBorder.Checked := CalcParam.Border;
  cbFog.Checked := PaintParam.Fog;

  DrawModeRadio[PaintParam.DrawMode].Checked := True;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.rsValueChanged(Sender: TObject);
begin
  if not (Sender is TRangeSelector) then
    Exit;

  with TRangeSelector(Sender) do begin
    DimEdit[Tag,2].Text := FloatToStr(Coordinates.Start[Tag] +
                            (Coordinates.Stop[Tag] - Coordinates.Start[Tag]) * LowValue / 1000);
    DimEdit[Tag,3].Text := FloatToStr(Coordinates.Start[Tag] +
                            (Coordinates.Stop[Tag] - Coordinates.Start[Tag]) * HighValue / 1000);
  end;
  Invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.raeLowNewChange(Sender: TObject);
begin
  if not (Sender is TCustomEdit) then
    Exit;

  try
    Coordinates.NewStart[TCustomEdit(Sender).Tag] :=
       StrToFloat(DimEdit[TCustomEdit(Sender).Tag, 2].Text);
    Invalidate;
  except
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.raeHighNewChange(Sender: TObject);
begin
  if not (Sender is TCustomEdit) then
    Exit;

  try
    Coordinates.NewStop[TCustomEdit(Sender).Tag] :=
      StrToFloat(DimEdit[TCustomEdit(Sender).Tag, 3].Text);
    Invalidate;
  except
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.BBtnResetClick(Sender: TObject);
begin
  SetDefaultStartStop;
  Coordinates.RecalcCoordTrans;
  FillControls;
  Calculate;
  Invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.bbtnZoomClick(Sender: TObject);
var
  i: Integer;
begin
  for i:= 0 to 2 do begin
    Coordinates.Start[i] := Coordinates.NewStart[i];
    Coordinates.Stop[i]  := Coordinates.NewStop[i];
  end;
  Coordinates.RecalcCoordTrans;
  FillControls;
  Calculate;
  Invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.btnRepaintClick(Sender: TObject);
var
  i: integer;
begin
  if CalcParam.CalcType <> cmbType.ItemIndex then begin
    CalcParam.CalcType := cmbType.ItemIndex;
    SetDefaultStartStop;
  end;

  for i:= 0 to 3 do begin
    try
      CalcParam.Param[i] := StrToFloat(ParamEdit[i].Text)
    except
      CalcParam.Param[i] := 0;
    end;
  end;

  try
    CalcParam.NrInter := StrToInt(raeNrInter.Text);
  except
    CalcParam.NrInter := 10;
  end;

  try
    Coordinates.GridSize := StrToInt(raeGridSize.Text);
  except
    Coordinates.GridSize := 10;
  end;

  try
    PaintParam.QuickDrawSize := StrToInt(raeQuickView.Text);
  except
    PaintParam.QuickDrawSize := 10000;
  end;

  CalcParam.Border := CbBoolBorder.Checked;
  PaintParam.Fog    := cbFog.Checked;

  for i:= 0 to 4 do begin
    if DrawModeRadio[i].Checked then
      PaintParam.DrawMode := i;
  end;

  if Sender = BbtnViewSettingsRepaint then
    WriteViewSettingsToIni;

  Coordinates.RecalcCoordTrans;
  Calculate;
  invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormActivate(Sender: TObject);
begin
  OnActivate := nil;

  Application.ProcessMessages;
  BbtnRandomClick(nil);
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.pnlObjectColorClick(Sender: TObject);
begin
  with TColorDialog.Create(nil) do begin
    Color := pnlObjectColor.Color;
    if Execute then begin
      pnlObjectColor.Color := Color;
      pnlObjectColor.Repaint;
      PaintParam.ObjectColor := ColorToRGB(Color);
      Self.Invalidate;
    end;
    Free;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.PageControlChange(Sender: TObject);
begin
  Invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.cbFogClick(Sender: TObject);
begin
  PaintParam.Fog := cbFog.Checked;
  invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.rbLightClick(Sender: TObject);
var
  i: Integer;
begin
  for i:= 0 to 4 do begin
    if DrawModeRadio[i].Checked then
      PaintParam.DrawMode := i;
  end;

  invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.SetDefaultStartStop;
var
  i: integer;
begin
  case CalcParam.CalcType of
   0,1:
     for i:= 0 to 2 do begin
       Coordinates.Start[i] := -2;
       Coordinates.Stop[i]  := 2;
     end;
   else
     for i:= 0 to 2 do begin
       Coordinates.Start[i] := -8;
       Coordinates.Stop[i]  := 8;
     end;
   end;
   for i:= 0 to 2 do begin
     Coordinates.NewStart[i] := Coordinates.Start[i];
     Coordinates.NewStop[i]  := Coordinates.Stop[i];
   end;
   FillControls;
   Coordinates.RecalcCoordTrans;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.pnlBackColorClick(Sender: TObject);
begin
  with TColorDialog.Create(nil) do begin
    Color := pnlBackColor.Color;
    if Execute then begin
      pnlBackColor.Color := Color;
      pnlBackColor.Repaint;
      PaintParam.BackColor := ColorToRGB(Color);
      Self.Invalidate;
    end;
    Free;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormShow(Sender: TObject);
begin
  OnShow := nil;
  ReadScreenPosFromIni;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.BbtnRandomClick(Sender: TObject);
begin
  if CalcParam.CalcType <> cmbType.ItemIndex then begin
    CalcParam.CalcType := cmbType.ItemIndex;
    SetDefaultStartStop;
  end;

  try
    CalcParam.NrInter := StrToInt(raeNrInter.Text);
  except
    CalcParam.NrInter := 10;
  end;

  CalcParam.Param[0] := 2 * Random - 1;
  CalcParam.Param[1] := 2 * Random - 1;
  CalcParam.Param[2] := 2 * Random - 1;
  CalcParam.Param[3] := 2 * Random - 1;
  FillControls;

  Calculate;
  invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.raeQuickViewChange(Sender: TObject);
begin
  try
    PaintParam.QuickDrawSize := StrToInt(raeQuickView.Text);
  except
  end;
end;

///////////////////////////////////////////////////////////////////////////////
// Camera handling
///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.RotateCamera(Right, Up, Turn: double);
var
  T: TVector3;
  sa, ca: double;
begin
  if Right <> 0 then begin
    sa := sin(Right);
    ca := cos(Right);
    t[0] := ca * PaintParam.ViewLook[0] + sa * PaintParam.ViewRight[0];
    t[1] := ca * PaintParam.ViewLook[1] + sa * PaintParam.ViewRight[1];
    t[2] := ca * PaintParam.ViewLook[2] + sa * PaintParam.ViewRight[2];

    PaintParam.ViewRight[0] := ca * PaintParam.ViewRight[0] - sa * PaintParam.ViewLook[0];
    PaintParam.ViewRight[1] := ca * PaintParam.ViewRight[1] - sa * PaintParam.ViewLook[1];
    PaintParam.ViewRight[2] := ca * PaintParam.ViewRight[2] - sa * PaintParam.ViewLook[2];

    PaintParam.ViewLook[0] := t[0];
    PaintParam.ViewLook[1] := t[1];
    PaintParam.ViewLook[2] := t[2];
  end;

  if Up <> 0 then begin
    sa := sin(Up);
    ca := cos(Up);

    t[0] := ca * PaintParam.ViewLook[0] + sa * PaintParam.Viewup[0];
    t[1] := ca * PaintParam.ViewLook[1] + sa * PaintParam.Viewup[1];
    t[2] := ca * PaintParam.ViewLook[2] + sa * PaintParam.Viewup[2];

    PaintParam.Viewup[0] := ca * PaintParam.Viewup[0] - sa * PaintParam.ViewLook[0];
    PaintParam.Viewup[1] := ca * PaintParam.Viewup[1] - sa * PaintParam.ViewLook[1];
    PaintParam.Viewup[2] := ca * PaintParam.Viewup[2] - sa * PaintParam.ViewLook[2];

    PaintParam.ViewLook[0] := t[0];
    PaintParam.ViewLook[1] := t[1];
    PaintParam.ViewLook[2] := t[2];
  end;

  if Turn <> 0 then begin
    sa := sin(Turn);
    ca := cos(Turn);

    t[0] := ca * PaintParam.ViewRight[0] + sa * PaintParam.Viewup[0];
    t[1] := ca * PaintParam.ViewRight[1] + sa * PaintParam.Viewup[1];
    t[2] := ca * PaintParam.ViewRight[2] + sa * PaintParam.Viewup[2];

    PaintParam.Viewup[0] := ca * PaintParam.Viewup[0] - sa * PaintParam.ViewRight[0];
    PaintParam.Viewup[1] := ca * PaintParam.Viewup[1] - sa * PaintParam.ViewRight[1];
    PaintParam.Viewup[2] := ca * PaintParam.Viewup[2] - sa * PaintParam.ViewRight[2];

    PaintParam.ViewRight[0] := t[0];
    PaintParam.ViewRight[1] := t[1];
    PaintParam.ViewRight[2] := t[2];
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Not Dragging then
    Exit;

  Dragging := False;

(*
  PaintParam.AngleHeight := PaintParam.AngleHeight + Y - MousePos.Y;
  if PaintParam.AngleHeight > 90 then
    PaintParam.AngleHeight := 90;
  if PaintParam.AngleHeight <  -90 then
    PaintParam.AngleHeight := -90;

  PaintParam.Azimuth := PaintParam.Azimuth + X - MousePos.X;
  if PaintParam.Azimuth > 360 then
    PaintParam.Azimuth := PaintParam.Azimuth - 360;
  if PaintParam.Azimuth <  -360 then
    PaintParam.Azimuth := PaintParam.Azimuth + 360;

  MousePos.X := X;
  MousePos.Y := Y;
*)
  Invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if Shift * [ssLeft, ssRight] = [] then
    Exit;
(*
  PaintParam.AngleHeight := PaintParam.AngleHeight + Y - MousePos.Y;
  if PaintParam.AngleHeight > 90 then
    PaintParam.AngleHeight := 90;
  if PaintParam.AngleHeight <  -90 then
    PaintParam.AngleHeight := -90;

  PaintParam.Azimuth := PaintParam.Azimuth + X - MousePos.X;
  if PaintParam.Azimuth > 360 then
    PaintParam.Azimuth := PaintParam.Azimuth - 360;
  if PaintParam.Azimuth <  -360 then
    PaintParam.Azimuth := PaintParam.Azimuth + 360;
*)
  if Shift = [ssleft] then
    RotateCamera((MousePos.X - X) * 0.02, (Y - MousePos.Y) * 0.02, 0)
  else
    RotateCamera(0,0, (X - MousePos.X ) * 0.02);


  MousePos.X := X;
  MousePos.Y := Y;

  Invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Shift * [ssLeft, ssRight] = [] then
    Exit;

  MousePos.X := X;
  MousePos.Y := Y;
  Dragging := True;
end;

///////////////////////////////////////////////////////////////////////////////
// Ini-functions
///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.ReadViewSettingsFromIni;
begin
  with TIniFile.Create(IniFile) do begin
    Coordinates.GridSize     := ReadInteger('Settings', 'GridSize'     , 40);
    PaintParam.ObjectColor   := ReadInteger('Settings', 'ObjectColor'  , RGB(127, 127, 255));
    PaintParam.BackColor     := ReadInteger('Settings', 'BackColor'    , RGB(0, 0, 0));
    CalcParam.Border         := ReadBool   ('Settings', 'Border'       , False);
    PaintParam.Fog           := ReadBool   ('Settings', 'Fog'          , True);
    PaintParam.DrawMode      := ReadInteger('Settings', 'DrawMode'     , 3);
    PaintParam.QuickDrawSize := ReadInteger('Settings', 'QuickDrawSize', 10000);
    Free;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.WriteViewSettingsToIni;
begin
  with TIniFile.Create(IniFile) do begin
    WriteInteger('Settings', 'GridSize'     , Coordinates.GridSize);
    WriteInteger('Settings', 'ObjectColor'  , PaintParam.ObjectColor);
    WriteInteger('Settings', 'BackColor'    , PaintParam.BackColor);
    WriteBool   ('Settings', 'Border'       , CalcParam.Border);
    WriteBool   ('Settings', 'Fog'          , PaintParam.Fog);
    WriteInteger('Settings', 'DrawMode'     , PaintParam.DrawMode);
    WriteInteger('Settings', 'QuickDrawSize', PaintParam.QuickDrawSize);
    Free;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.ReadScreenPosFromIni;
begin
  with TIniFile.Create(IniFile) do begin
    Left   := ReadInteger('ScreenPos', 'Left',   (Screen.Width - Width) div 2);
    Top    := ReadInteger('ScreenPos', 'Top',    (Screen.Height - Height) div 2);
    Width  := ReadInteger('ScreenPos', 'Width',  Width);
    Height := ReadInteger('ScreenPos', 'Height', Height);
    Free;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.WriteScreenPosToIni;
begin
  if WindowState <> wsNormal then
    Exit;

  with TIniFile.Create(IniFile) do begin
    WriteInteger('ScreenPos', 'Width',  Width);
    WriteInteger('ScreenPos', 'Height', Height);
    WriteInteger('ScreenPos', 'Left',   Left);
    WriteInteger('ScreenPos', 'Top',    Top);
    Free;
  end;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.tbLoadChange(Sender: TObject);
var
  section: string;
begin
  section := 'Save_' + IntToStr(tbLoad.Position);

  with TIniFile.Create(IniFile) do begin
    CalcParam.CalcType := ReadInteger(section, 'CalcType', 0);
    CalcParam.Param[0] := ReadFloat(section, 'Param0', 0);
    CalcParam.Param[1] := ReadFloat(section, 'Param1', 0);
    CalcParam.Param[2] := ReadFloat(section, 'Param2', 0);
    CalcParam.Param[3] := ReadFloat(section, 'Param3', 0);
    CalcParam.NrInter  := ReadInteger(section, 'NrInter', 10);
    SetDefaultStartStop;

    Coordinates.Start[0] := ReadFloat(section, 'Start0', -2);
    Coordinates.Start[1] := ReadFloat(section, 'Start1', -2);
    Coordinates.Start[2] := ReadFloat(section, 'Start2', -2);
    Coordinates.Stop[0]  := ReadFloat(section, 'Stop0', 2);
    Coordinates.Stop[1]  := ReadFloat(section, 'Stop1', 2);
    Coordinates.Stop[2]  := ReadFloat(section, 'Stop2', 2);

    PaintParam.Viewlook[0] := ReadFloat(section, 'camlook0', 1);
    PaintParam.Viewlook[1] := ReadFloat(section, 'camlook1', 0);
    PaintParam.Viewlook[2] := ReadFloat(section, 'camlook2', 0);
    PaintParam.ViewUp[0] := ReadFloat(section, 'camup0', 0);
    PaintParam.ViewUp[1] := ReadFloat(section, 'camup1', 0);
    PaintParam.ViewUp[2] := ReadFloat(section, 'camup2', 1);
    PaintParam.ViewRight[0] := ReadFloat(section, 'camRight0', 0);
    PaintParam.ViewRight[1] := ReadFloat(section, 'camRight1', 1);
    PaintParam.ViewRight[2] := ReadFloat(section, 'camRight2', 0);

    Free;
  end;

  Coordinates.RecalcCoordTrans;
  FillControls;
  Calculate;
  invalidate;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.GetNrSave;
begin
  with TIniFile.Create(IniFile) do begin
    NrSaves := ReadInteger('Saves', 'NrSaves', 0);
    Free;
  end;

  tbLoad.Visible := NrSaves <> 0;
  if NrSaves > 0 then
    tbLoad.Max := NrSaves - 1;
end;

///////////////////////////////////////////////////////////////////////////////
procedure TfrmMain.btnSaveClick(Sender: TObject);
var
  section: string;
begin
  section := 'Save_' + IntToStr(NrSaves);

  with TIniFile.Create(IniFile) do begin
    WriteInteger(section, 'CalcType', CalcParam.CalcType);
    WriteFloat(section, 'Param0', CalcParam.Param[0]);
    WriteFloat(section, 'Param1', CalcParam.Param[1]);
    WriteFloat(section, 'Param2', CalcParam.Param[2]);
    WriteFloat(section, 'Param3', CalcParam.Param[3]);
    WriteInteger(section, 'NrInter', CalcParam.NrInter);

    WriteFloat(section, 'Start0', Coordinates.Start[0]);
    WriteFloat(section, 'Start1', Coordinates.Start[1]);
    WriteFloat(section, 'Start2', Coordinates.Start[2]);
    WriteFloat(section, 'Stop0', Coordinates.Stop[0]);
    WriteFloat(section, 'Stop1', Coordinates.Stop[1]);
    WriteFloat(section, 'Stop2', Coordinates.Stop[2]);

    WriteFloat(section, 'camlook0', PaintParam.Viewlook[0]);
    WriteFloat(section, 'camlook1', PaintParam.Viewlook[1]);
    WriteFloat(section, 'camlook2', PaintParam.Viewlook[2]);
    WriteFloat(section, 'camup0', PaintParam.ViewUp[0]);
    WriteFloat(section, 'camup1', PaintParam.ViewUp[1]);
    WriteFloat(section, 'camup2', PaintParam.ViewUp[2]);
    WriteFloat(section, 'camRight0', PaintParam.ViewRight[0]);
    WriteFloat(section, 'camRight1', PaintParam.ViewRight[1]);
    WriteFloat(section, 'camRight2', PaintParam.ViewRight[2]);

    Inc(NrSaves);
    WriteInteger('Saves', 'NrSaves', NrSaves);

    Free;
  end;

  tbLoad.Visible := NrSaves <> 0;
  if NrSaves > 0 then
    tbLoad.Max := NrSaves - 1;
end;


///////////////////////////////////////////////////////////////////////////////
end.
